from typing import Callable, Any, Dict

import dict_tools.differ as differ

async def run(
    hub,
    state_name:str,
    state_describe_func: Callable,
    state_describe_kwargs: Dict[str, Any],
    state_func: Callable,
    state_func_kwargs: Dict[str, Any],
    fail_on_error_before:bool = None,
    fail_on_error_after:bool = None,
    succeed_on_error_before:bool = None,
    succeed_on_error_after:bool = None,
    succeed_on_describe_before:bool=False,
    fail_on_describe_after:bool=False,
):
    assert not (fail_on_error_before is not None and succeed_on_error_before is not None), "Mutually exclusive args"
    assert not (fail_on_error_after is not None and succeed_on_error_after is not None), "Mutually exclusive args"
    try:
        before = state_describe_func(**state_describe_kwargs)
        before = await hub.pop.loop.unwrap(before)
        if before:
            raise ChildProcessError(f"Resource exists")
        if succeed_on_describe_before and before:
            return {
                "comment": f"{state_name} already complete",
                "changes": {},
                "name": state_name,
                "result": True,
            }
    except Exception as e:
        if fail_on_error_before:
            return {
                "comment": f"{e.__class__.__name__}: {e}",
                "changes": {},
                "name": state_name,
                "result": False,
            }
        elif succeed_on_error_before:
            return {
                "comment": f"{state_name} already complete: {e.__class__.__name__}: {e}",
                "changes": {},
                "name": state_name,
                "result": True,
            }
        else:
            before = {}
            comment = f"{e.__class__.__name__} while describing resource {state_describe_func} before {state_func}: {e}"

    assert isinstance(before, dict), f"Return from describe function '{state_describe_func}' is not a dict"

    try:
        ret = state_func(**state_func_kwargs)
        ret = await hub.pop.loop.unwrap(ret)
    except Exception as e:
        return {
            "comment": f"{e.__class__.__name__}: {e}",
            "changes": {},
            "name": state_name,
            "result": False,
        }

    try:
        after = state_describe_func(**state_describe_kwargs)
        after = await hub.pop.loop.unwrap(after)
        if not after:
            raise ChildProcessError(f"Resource does not exist")
        if fail_on_describe_after and after:
            return {
                "comment": comment,
                "changes": differ.deep_diff(before, after),
                "name": state_name,
                "result": False,
            }
    except Exception as e:
        if fail_on_error_after:
            return {
                "comment": f"{e.__class__.__name__}: {e}",
                "changes": {},
                "name": state_name,
                "result": False,
            }
        elif succeed_on_error_after:
            return {
                "comment": f"{state_name} completed: {e.__class__.__name__}: {e}",
                "changes": differ.deep_diff(before, {}),
                "name": state_name,
                "result": True,
            }
        else:
            after = {}
            comment = f"{e.__class__.__name__} while describing resource {state_describe_func} after {state_func}: {e}"

    return {
        "comment": comment,
        "changes": differ.deep_diff(before, after),
        "name": state_name,
        "result": True,
    }



def seq(hub, low, running):
    """
    Process the multi stage routine to determine the current requisite sequence
    """
    # TODO: Make this more pluggable. This means that plugins get detected and run,
    # Granted, we only need to do this for prereq, I don't think that we will need
    # to add another plugin here.
    ret = hub.idem.req.seq.straight.run(low, running)
    ret = hub.idem.req.seq.prereq.run(ret, low, running)
    return ret



from colored import fore

from idem.exec.init import ExecReturn


def display(hub, data):
    """
    Display the data from an idem run
    """
    if not isinstance(data, ExecReturn):
        return hub.output.nested.display(data)

    if not data.result:
        return fore.RED + str(data.comment)
    else:
        return hub.output.nested.display(data.ret)



import dict_tools.data as data


async def start(hub, name):
    """
    Called only after the named run has compiled low data. If no low data
    is present an exception will be raised
    """
    if not hub.idem.RUNS[name].get("low"):
        raise ValueError(f"No low data for '{name}' in RUNS")
    ctx = data.NamespaceDict({"run_name": name, "test": hub.idem.RUNS[name]["test"]})
    rtime = hub.idem.RUNS[name]["runtime"]
    low = hub.idem.RUNS[name].get("low")
    ref = f"idem.run.{rtime}.runtime"
    old_seq = {}
    old_seq_len = -1
    needs_post_low = True
    while True:
        seq = hub.idem.req.seq.init.seq(low, hub.idem.RUNS[name]["running"])
        if seq == old_seq:
            unmet_reqs = {chunk["tag"]: chunk["unmet"] for chunk in seq.values()}
            raise Exception(f"No sequence changed for '{name}': {unmet_reqs}")
        await getattr(hub, ref)(name, ctx, seq, low, hub.idem.RUNS[name]["running"])
        await hub.idem.resolve.render(name)
        await hub.idem.state.compile(name)
        low = hub.idem.RUNS[name].get("low")
        if len(low) <= len(hub.idem.RUNS[name]["running"]):
            if hub.idem.RUNS[name]["post_low"] and needs_post_low:
                hub.idem.RUNS[name]["low"].extend(hub.idem.RUNS[name]["post_low"])
                needs_post_low = False
                continue
            else:
                break
        if len(seq) == old_seq_len:
            raise RecursionError(f"No progress made on '{name}', Recursive Requisite!")
        old_seq = seq
        old_seq_len = len(seq)
        # Check for any new, available blocks to render



import asyncio
import inspect
import warnings
from typing import Awaitable

from idem.exec.init import ExecReturn


def _create_exec_return(ret, ref: str):
    try:
        return ExecReturn(
            **ret,
            ref=ref,
        )
    except TypeError:
        # TODO For now, log a warning when exec module output doesn't conform to this format.
        #  later, remove this try/except
        warnings.warn(
            "Exec modules must return a dictionary {'result': True|False, 'comment': Any, 'ret': Any}",
            DeprecationWarning,
        )
        return ret


async def _create_exec_return_coro(ret: Awaitable, ref: str):
    return _create_exec_return(await ret, ref)


def post(hub, ctx):
    """
    Convert the dict return to an immutable namespace addressable format
    """
    # TODO this isn't happening
    ref = ctx.func.__module__
    if asyncio.iscoroutine(ctx.ret):
        return _create_exec_return_coro(ctx.ret, ref)
    else:
        return _create_exec_return(ctx.ret, ref)


def _format_error(hub, ctx, exc: Exception):
    ret_str = f"{exc.__class__.__name__}: {inspect.getfile(ctx.func)}:{ctx.func.__name__}: {exc}"
    hub.log.error(ret_str)
    return {"result": False, "ret": None, "comment": ret_str}



# NOTES
# This is a VERY simple outputter for idem, it does not do everything the
# Salt highstate outputter does, and nor should it! This outputter should
# not become hyper complicated, things like terse should be another
# outputter, this should really just get things like errors added
from colored import attr
from colored import fg


def display(hub, data):
    """
    Display the data from an idem run
    """
    if not isinstance(data, dict):
        return hub.output.nested.display(data)
    endc = attr(0)
    strs = []
    for tag in sorted(data, key=lambda k: data[k].get("__run_num", 0)):
        ret = data[tag]
        comps = tag.split("_|-")
        state = comps[0]
        id_ = comps[1]
        fun = comps[3]
        name = ret["name"]
        result = ret["result"]
        comment = ret["comment"]
        changes = hub.output.nested.display(ret["changes"])
        if result is True and changes:
            tcolor = fg(6)
        elif result is True:
            tcolor = fg(2)
        elif result is None:
            tcolor = fg(11)
        elif result is False:
            tcolor = fg(9)

        strs.append(f"{tcolor}--------{endc}")
        strs.append(f"{tcolor}      ID: {id_}{endc}")
        strs.append(f"{tcolor}Function: {state}.{fun}{endc}")
        strs.append(f"{tcolor}  Result: {result}{endc}")
        strs.append(f"{tcolor} Comment: {comment}{endc}")
        strs.append(f"{tcolor} Changes: {changes}{endc}")
    return "\n".join(strs)



def req_map(hub):
    """
    Gather the requisite restrtrictions and populate the requisite behavior map
    """
    rmap = {}
    for mod in hub.idem.req:
        if mod.__name__ == "init":
            continue
        if hasattr(mod, "define"):
            rmap[mod.__name__] = mod.define()
    hub.idem.RMAP = rmap



import copy


def check(hub, name, ctx, condition, req_ret, chunk):
    """
    If changes are made then run the configured post command
    """
    ret = {}
    if req_ret["ret"]["changes"]:
        id_ = f"{chunk['__id__']}_listen"
        post_chunk = copy.deepcopy(chunk)
        post_chunk["fun"] = condition
        post_chunk["__id__"] = id_
        post_chunk["name"] = id_
        post_chunk["order"] = -1
        hub.idem.RUNS[name]["post_low"].append(post_chunk)
    return ret



# The order of the sequence that needs to be implemented:
# Start with a single sls file, just like you started with salt
# Stub out the routines around gathering the initial sls file
# Just use a yaml renderer and get it to where we can manage some basic
# includes to drive to highdata
# Then we can start to fill out renderers while at the same time
# deepening the compiler
import pathlib
import sys

from idem.exec.init import ExecReturn


def __init__(hub):
    hub.pop.sub.load_subdirs(hub.idem, recurse=True)
    hub.idem.RUNS = {}
    hub.pop.sub.add(dyne_name="log")
    hub.pop.sub.add(dyne_name="acct")
    hub.pop.sub.add(dyne_name="rend")
    hub.pop.sub.add(dyne_name="output")
    # hub.pop.sub.add(dyne_name="cache")
    hub.pop.sub.add(dyne_name="tool")
    hub.pop.sub.load_subdirs(hub.tool, recurse=True)
    hub.pop.sub.add(dyne_name="exec")
    hub.pop.sub.load_subdirs(hub.exec, recurse=True)
    hub.pop.sub.add(dyne_name="states")
    hub.pop.sub.load_subdirs(hub.states, recurse=True)
    hub.idem.req.init.req_map()


def cli(hub):
    """
    Execute a single idem run from the cli
    """
    hub.pop.config.load(["idem", "acct"], cli="idem")
    # hub.cache.PLUGIN = hub.OPT.idem.cache_plugin
    hub.pop.loop.create()
    retcode = hub.pop.Loop.run_until_complete(hub.idem.init.cli_apply())
    sys.exit(retcode)


# If the gathering and cli def funcs grow they should be moved to a plugin
def get_refs(hub):
    """
    Determine where the sls sources are
    """
    sls_sources = []
    slses = []
    if hub.OPT.idem.tree:
        tree = f"file://{hub.OPT.idem.tree}"
        sls_sources.append(tree)
    for sls in hub.OPT.idem.sls:
        path = pathlib.Path(sls)
        if path.is_file():
            ref = str(path.stem if path.suffix == ".sls" else path.name)
            slses.append(ref)
            implied = f"file://{path.parent}"
            if implied not in sls_sources:
                sls_sources.append(implied)
        else:
            slses.append(sls)

    sls_sources.extend(hub.OPT["idem"]["sls_sources"])

    return {"sls_sources": sls_sources, "sls": slses}


async def cli_apply(hub):
    """
    Run the CLI routine in a loop
    """
    retcode = 0

    try:
        if hub.SUBPARSER == "state":
            retcode = await hub.idem.init.cli_sls()
        elif hub.SUBPARSER == "exec":
            retcode = await hub.idem.init.cli_exec()
    finally:
        if hub.acct.UNLOCKED:
            await hub.acct.init.close()

    return retcode


async def cli_sls(hub) -> int:
    """
    Execute the cli routine to run states
    """
    src = hub.idem.init.get_refs()
    name = "cli"
    await hub.idem.state.apply(
        name=name,
        sls_sources=src["sls_sources"],
        render=hub.OPT.idem.render,
        runtime=hub.OPT.idem.runtime,
        subs=["states"],
        cache_dir=hub.OPT.idem.cache_dir,
        sls=src["sls"],
        test=hub.OPT.idem.test,
        acct_file=hub.OPT.acct.acct_file,
        acct_key=hub.OPT.acct.acct_key,
        acct_profile=hub.OPT.idem.acct_profile,
    )

    errors = hub.idem.RUNS[name]["errors"]
    if errors:
        display = hub.output.nested.display(errors)
        print(display)
        # Return a non-zero error code
        return len(errors)
    running = hub.idem.RUNS[name]["running"]
    output = hub.OPT.idem.output or "state"
    display = hub.output[output].display(running)
    print(display)
    return 0


async def cli_exec(hub) -> int:
    exec_path = hub.OPT.idem.exec_func
    exec_args = hub.OPT.idem.exec_args
    if not exec_path.startswith("exec"):
        exec_path = f"exec.{exec_path}"
    args = []
    kwargs = {}
    for arg in exec_args:
        if isinstance(arg, dict):
            kwargs.update(arg)
        else:
            args.append(arg)
    ret = await hub.idem.ex.run(
        exec_path,
        args,
        kwargs,
        hub.OPT.acct.acct_file,
        hub.OPT.acct.acct_key,
        hub.OPT.idem.acct_profile,
    )

    output = hub.OPT.idem.output or "exec"
    display = hub.output[output].display(ret)
    print(display)

    if isinstance(ret, ExecReturn):
        return int(not ret.result)

    return 1



def resolve(hub, rdats):
    """
    Given the dict of requisite data sets, determine if the defined
    requisites passed. This allows for requisites to be defined
    as requiring, all, some, one, or none of the required states to
    pass
    """
    errors = []
    for req, rlist in rdats.items():
        rmap = hub.idem.RMAP[req]
        resolver = rmap.get("resolver", "all")
        errors.extend(getattr(hub, f"idem.resolver.{resolver}.check")(rlist))
    return errors



import pop.loader


# These are keywords passed to state module functions which are to be used
# by salt in this state module and not on the actual state module function
STATE_REQUISITE_KEYWORDS = frozenset(
    [
        "onchanges",
        "onchanges_any",
        "onfail",
        "onfail_any",
        "onfail_all",
        "onfail_stop",
        "prereq",
        "prerequired",
        "watch",
        "watch_any",
        "require",
        "require_any",
        "listen",
    ]
)
STATE_REQUISITE_IN_KEYWORDS = frozenset(
    [
        "onchanges_in",
        "onfail_in",
        "prereq_in",
        "watch_in",
        "require_in",
        "listen_in",
    ]
)
STATE_RUNTIME_KEYWORDS = frozenset(
    [
        "fun",
        "state",
        "check_cmd",
        "failhard",
        "onlyif",
        "unless",
        "retry",
        "order",
        "parallel",
        "prereq",
        "prereq_in",
        "prerequired",
        "reload_modules",
        "reload_grains",
        "reload_pillar",
        "runas",
        "runas_password",
        "fire_event",
        "saltenv",
        "use",
        "use_in",
        "__run_name",
        "__env__",
        "__sls__",
        "__id__",
        "__orchestration_jid__",
        "__pub_user",
        "__pub_arg",
        "__pub_jid",
        "__pub_fun",
        "__pub_tgt",
        "__pub_ret",
        "__pub_pid",
        "__pub_tgt_type",
        "__prereq__",
    ]
)

STATE_INTERNAL_KEYWORDS = STATE_REQUISITE_KEYWORDS.union(
    STATE_REQUISITE_IN_KEYWORDS
).union(STATE_RUNTIME_KEYWORDS)


def get_func(hub, name, chunk, fun=None):
    """
    Given the runtime name and the chunk in question, determine what function
    on the hub that can be run
    """
    if fun is None:
        fun = chunk["fun"]
    s_ref = chunk["state"]
    for sub in hub.idem.RUNS[name]["subs"]:
        test = f"{sub}.{s_ref}.{fun}"
        try:
            func = getattr(hub, test)
        except AttributeError:
            continue
        if isinstance(func, pop.loader.LoadedMod):
            continue
        if func is None:
            continue
        return func
    return None


async def run(hub, name, ctx, low, seq_comp, running, run_num):
    """
    All requisites have been met for this low chunk.
    """
    chunk = seq_comp["chunk"]
    tag = hub.idem.tools.gen_tag(chunk)
    skip_check = ["resolver"]
    rdats = {}
    errors = []
    for reqret in seq_comp.get("reqrets", []):
        req = reqret["req"]
        if req not in rdats:
            rdats[req] = []
        rules = hub.idem.RMAP[req]
        for rule in rules:
            if rule in skip_check:
                continue
            if hasattr(hub.idem.rules, rule):
                rdat = getattr(hub.idem.rules, rule).check(
                    name, ctx, rules[rule], reqret, chunk
                )
                rdats[req].append(rdat)
    errors = hub.idem.resolver.init.resolve(rdats)
    if errors:
        running[tag] = {
            "name": chunk["name"],
            "changes": {},
            "comment": "\n".join(errors),
            "result": False,
            "__run_num": run_num,
        }
        return
    func = hub.idem.rules.init.get_func(name, chunk)
    if func is None:
        running[tag] = {
            "name": chunk["name"],
            "changes": {},
            "comment": f'The named state {chunk["state"]} is not available',
            "result": False,
            "__run_num": run_num,
        }
        return
    chunk["ctx"] = ctx
    chunk = await hub.idem.mod.init.modify(name, chunk)
    call = hub.idem.tools.format_call(
        func, chunk, expected_extra_kws=STATE_INTERNAL_KEYWORDS
    )
    for req, rlist in rdats.items():
        for rdat in rlist:
            if "pre" in rdat:
                ret = rdat["pre"](*call["args"], **call["kwargs"])
                await hub.pop.loop.unwrap(ret)
    ret = func(*call["args"], **call["kwargs"])
    ret = await hub.pop.loop.unwrap(ret)
    for req, rlist in rdats.items():
        for rdat in rlist:
            if "post" in rdat:
                ret = rdat["post"](*call["args"], **call["kwargs"])
                ret = await hub.pop.loop.unwrap(ret)
    ret["__run_num"] = run_num
    running[tag] = ret



def run(hub, low, running):
    """
    Return the sequence map that should be used to execute the lowstate
    The sequence needs to identify:
    1. recursive requisites
    2. what chunks are free to run
    3. Behavior augments for the next chunk to run
    """
    ret = {}
    for ind, chunk in enumerate(low):
        tag = hub.idem.tools.gen_tag(chunk)
        if tag in running:
            # Already ran this one, don't add it to the sequence
            continue
        ret[ind] = {"chunk": chunk, "reqrets": [], "unmet": set(), "tag": tag}
        for req, data in hub.idem.RMAP.items():
            if data.get("prereq"):
                continue
            if req in chunk:
                for rdef in chunk[req]:
                    if not isinstance(rdef, dict):
                        # TODO: Error check
                        continue
                    state = next(iter(rdef))
                    name = rdef[state]
                    r_chunks = hub.idem.tools.get_chunks(low, state, name)
                    if not r_chunks:
                        ret[ind]["errors"].append(
                            f"Requisite {req} {state}:{name} not found"
                        )
                    for r_chunk in r_chunks:
                        r_tag = hub.idem.tools.gen_tag(r_chunk)
                        if r_tag in running:
                            reqret = {
                                "req": req,
                                "name": name,
                                "state": state,
                                "r_tag": r_tag,
                                "ret": running[r_tag],
                                "chunk": "r_chunk",
                            }
                            # it has been run, check the rules
                            ret[ind]["reqrets"].append(reqret)
                        else:
                            ret[ind]["unmet"].add(r_tag)
    return ret



"""
Test States
===========

Provide test case states that enable easy testing of things to do with state
calls, e.g. running, calling, logging, output filtering etc.

.. code-block:: yaml

    always-passes-with-any-kwarg:
      test.nop:
        - name: foo
        - something: else
        - foo: bar

    always-passes:
      test.succeed_without_changes:
        - name: foo

    always-fails:
      test.fail_without_changes:
        - name: foo

    always-changes-and-succeeds:
      test.succeed_with_changes:
        - name: foo

    always-changes-and-fails:
      test.fail_with_changes:
        - name: foo
"""
import random

TREQ = {
    "treq": {
        "require": [
            "test.nop",
        ]
    },
}


def treq(hub, ctx, name, **kwargs):
    """
    Ensure that a transparent requisite is applied
    """
    return succeed_without_changes(hub, ctx, name)


def nop(hub, ctx, name, **kwargs):
    """
    A no-op state that does nothing. Useful in conjunction with the `use`
    requisite, or in templates which could otherwise be empty due to jinja
    rendering
    """
    if ctx["test"]:
        return none_without_changes(hub, ctx, name)
    return succeed_without_changes(hub, ctx, name)


async def anop(hub, ctx, name, **kwargs):
    """
    A no-op state that does nothing. Useful in conjunction with the `use`
    requisite, or in templates which could otherwise be empty due to jinja
    rendering
    """
    if ctx["test"]:
        return none_without_changes(hub, ctx, name)
    return succeed_without_changes(hub, ctx, name)


def succeed_with_comment(hub, ctx, name, comment, **kwargs):
    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": comment,
    }
    return ret


def succeed_without_changes(hub, ctx, name, **kwargs):
    """
    name
        A unique string.
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": "Success!"}
    return ret


def none_without_changes(hub, ctx, name, **kwargs):
    """
    name
        A unique string.
    """
    ret = {"name": name, "changes": {}, "result": None, "comment": "Success!"}
    return ret


def fail_without_changes(hub, ctx, name, **kwargs):
    """
    Returns failure.

    name:
        A unique string.
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": "Failure!"}

    return ret


def succeed_with_changes(hub, ctx, name, **kwargs):
    """
    Returns successful and changes is not empty

    name:
        A unique string.
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": "Success!"}

    ret["changes"] = {
        "testing": {"old": "Unchanged", "new": "Something pretended to change"}
    }

    return ret


def fail_with_changes(hub, ctx, name, **kwargs):
    """
    Returns failure and changes is not empty.

    name:
        A unique string.
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": "Failure!"}
    ret["changes"] = {
        "testing": {"old": "Unchanged", "new": "Something pretended to change"}
    }
    return ret


def update_low(hub, ctx, name):
    """
    Use the __run_name to add a run to the low
    """
    extra = {
        "__sls__": "none",
        "name": "totally_extra_alls",
        "__id__": "king_arthur",
        "state": "test",
        "fun": "nop",
    }
    hub.idem.RUNS[ctx["run_name"]]["add_low"].append(extra)
    return succeed_without_changes(hub, ctx, name)


def mod_watch(hub, ctx, name, **kwargs):
    """
    Return a mod_watch call for test
    """
    ret = {
        "name": name,
        "changes": {"watch": True},
        "result": True,
        "comment": "Watch ran!",
    }
    return ret


def configurable_test_state(
    hub, ctx, name, changes=True, result=True, comment="", **kwargs
):
    """
    A configurable test state which determines its output based on the inputs.

    name:
        A unique string.
    changes:
        Do we return anything in the changes field?
        Accepts True, False, and 'Random'
        Default is True
    result:
        Do we return successfully or not?
        Accepts True, False, and 'Random'
        Default is True
        If test is True and changes is True, this will be None.  If test is
        True and and changes is False, this will be True.
    comment:
        String to fill the comment field with.
        Default is ''
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": comment}

    change_data = {
        "testing": {"old": "Unchanged", "new": "Something pretended to change"}
    }

    # If changes is True, then we place our dummy change dictionary into it
    if changes == "Random":
        if random.choice([True, False]):
            ret["changes"] = change_data
    elif changes is True:
        ret["changes"] = change_data
    elif changes is False:
        ret["changes"] = {}

    if result == "Random":
        ret["result"] = random.choice([True, False])
    elif result is True:
        ret["result"] = True
    elif result is False:
        ret["result"] = False

    if ctx["test"]:
        ret["result"] = True if changes is False else None
        ret["comment"] = "This is a test" if not comment else comment

    return ret



"""
The sls resolver is used to gather sls files, render them and return the initial
phase 1 highdata. This involves translating sls references into file paths,
downloading those sls files and then rendering them.

Once an sls file is rendered the include statements are resolved as well.
"""
import re


async def gather(hub, name, *sls):
    """
    Gather the named sls references
    """
    await hub.idem.resolve.get_blocks(name, sls)
    await hub.idem.resolve.render(name)


async def get_blocks(hub, name, sls):
    for sls_ref in sls:
        cfn = await hub.idem.get.ref(name, sls_ref)
        if not cfn:
            hub.idem.RUNS[name]["errors"].append(
                f"SLS ref {sls_ref} did not resolve to a file"
            )
            continue
        blocks = hub.rend.init.blocks(cfn)
        hub.idem.RUNS[name]["blocks"][sls_ref] = blocks
        hub.idem.RUNS[name]["sls_refs"][sls_ref] = cfn
        hub.idem.RUNS[name]["resolved"].add(sls_ref)
        hub.idem.RUNS[name]["files"].add(cfn)


async def render(hub, name):
    """
    Pop the available blocks and render them if they have satisfied requisites
    """
    rendered = {}
    for sls_ref, blocks in hub.idem.RUNS[name]["blocks"].items():
        cfn = hub.idem.RUNS[name]["sls_refs"][sls_ref]
        for bname, block in blocks.items():
            clear = True
            for key, val in block.get("keys", {}).items():
                # TODO: This should be an additional render requisite plugin
                # subsystem, change it to a subsystem as soon as any new conditionals
                # are added!!
                clear = False
                if key == "require":
                    for tag, data in hub.idem.RUNS[name]["running"].items():
                        if data["name"] == val:
                            clear = True
                            break
                if clear:
                    continue
            if clear:
                state = await hub.rend.init.parse_bytes(
                    block, hub.idem.RUNS[name]["render"]
                )
                await hub.idem.resolve.introduce(name, state, sls_ref, cfn)
                rendered[sls_ref] = bname
    for sls_ref, bname in rendered.items():
        hub.idem.RUNS[name]["blocks"][sls_ref].pop(bname)


async def introduce(hub, name, state, sls_ref, cfn):
    """
    Introduce the raw state into the running dataset
    """
    if not isinstance(state, dict):
        hub.idem.RUNS[name]["errors"].append(
            f"SLS {sls_ref} is not formed as a dict but as a {type(state)}"
        )
        return
    if "include" in state:
        if not isinstance(state["include"], list):
            hub.idem.RUNS[name]["errors"].append(
                f'Include Declaration in SLS {sls_ref} is not formed as a list but as a {type(state["include"])}'
            )
        include = state.pop("include")
    else:
        include = []
    hub.idem.resolve.extend(name, state, sls_ref)
    hub.idem.resolve.exclude(name, state, sls_ref)
    hub.idem.resolve.decls(name, state, sls_ref)
    hub.idem.resolve.iorder(name, state)
    await hub.idem.resolve.includes(name, include, state, sls_ref, cfn)
    hub.idem.RUNS[name]["high"].update(state)


def iorder(hub, name, state):
    """
    Take a state and apply the iorder system
    """
    for id_ in state:
        for s_dec in state[id_]:
            if not isinstance(s_dec, str):
                # PyDSL OrderedDict?
                continue

            if not isinstance(state[id_], dict):
                # Include's or excludes as lists?
                continue
            if not isinstance(state[id_][s_dec], list):
                # Bad syntax, let the verify seq pick it up later on
                continue

            found = False
            if s_dec.startswith("_"):
                continue

            for arg in state[id_][s_dec]:
                if isinstance(arg, dict):
                    if len(arg) > 0:
                        if next(iter(arg)) == "order":
                            found = True
            if not found:
                if not isinstance(state[id_][s_dec], list):
                    # quite certainly a syntax error, managed elsewhere
                    continue
                state[id_][s_dec].append({"order": hub.idem.RUNS[name]["iorder"]})
                hub.idem.RUNS[name]["iorder"] += 1


def extend(hub, name, state, sls_ref):
    """
    Resolve the extend statement
    """
    if "extend" in state:
        ext = state.pop("extend")
        if not isinstance(ext, dict):
            hub.idem.RUNS[name]["errors"].append(
                f'Extension value in SLS "{sls_ref}" is not a dictionary'
            )
            return
        for id_ in ext:
            if not isinstance(ext[id_], dict):
                hub.idem.RUNS[name]["errors"].append(
                    f'Extension ID "{id_}" in SLS "{sls_ref}" is not a dictionary'
                )
                continue
            if "__sls__" not in ext[id_]:
                ext[id_]["__sls__"] = sls_ref
            # if '__env__' not in ext[id_]:
            #    ext[id_]['__env__'] = saltenv
            for key in list(ext[id_]):
                if key.startswith("_"):
                    continue
                if not isinstance(ext[id_][key], list):
                    continue
                if "." in key:
                    comps = key.split(".")
                    ext[id_][comps[0]] = ext[id_].pop(key)
                    ext[id_][comps[0]].append(comps[1])
        state.setdefault("__extend__", []).append(ext)


def exclude(hub, name, state, sls_ref):
    """
    Resolve any exclude statements
    """
    if "exclude" in state:
        exc = state.pop("exclude")
        if not isinstance(exc, list):
            hub.idem.RUNS[name]["errors"].append(
                f"Exclude Declaration in SLS {sls_ref} is not formed as a list"
            )
        state.setdefault("__exclude__", []).extend(exc)


def decls(hub, name, state, sls_ref):
    """
    Resolve and state formatting and data insertion
    """
    for id_ in state:
        if not isinstance(state[id_], dict):
            if id_ == "__extend__":
                continue
            if id_ == "__exclude__":
                continue

            if isinstance(state[id_], str):
                # Is this is a short state, it needs to be padded
                if "." in state[id_]:
                    comps = state[id_].split(".")
                    state_ref = ".".join(comps[:-1])
                    state[id_] = {"__sls__": sls_ref, state_ref: [comps[-1]]}
                    continue
            hub.idem.RUNS[name]["errors"].append(
                f"ID {id_} in SLS {sls_ref} is not a dictionary"
            )
            continue
        skeys = set()
        for key in list(state[id_]):
            if key.startswith("_"):
                continue
            if not isinstance(state[id_][key], list):
                continue
            if "." in key:
                comps = key.split(".")
                # Idem doesn't support state files such as:
                #
                #     /etc/redis/redis.conf:
                #       file.managed:
                #         - user: redis
                #         - group: redis
                #         - mode: 644
                #       file.comment:
                #           - regex: ^requirepass
                ref = ".".join(comps[:-1])
                if ref in skeys:
                    hub.idem.RUNS[name]["errors"].append(
                        f'ID "{id_}" in SLS "{sls_ref}" contains multiple state declarations of the same type'
                    )
                    continue
                state[id_][ref] = state[id_].pop(key)
                state[id_][ref].append(comps[-1])
                skeys.add(ref)
                continue
            skeys.add(key)
        if "__sls__" not in state[id_]:
            state[id_]["__sls__"] = sls_ref


async def includes(hub, name, include, state, sls_ref, cfn):
    """
    Parse through the includes and download not-yet-resolved includes
    """
    for inc_sls in include:
        if inc_sls.startswith("."):
            match = re.match(r"^(\.+)(.*)$", inc_sls)
            if match:
                levels, include = match.groups()
            else:
                hub.idem.RUNS[name]["errors"].append(
                    f'Badly formatted include {inc_sls} found in SLS "{sls_ref}"'
                )
                continue
            level_count = len(levels)
            p_comps = sls_ref.split(".")
            if cfn.endswith("/init.sls"):
                p_comps.append("init")
            if level_count > len(p_comps):
                hub.idem.RUNS[name]["errors"].append(
                    f'Attempted relative include of "{inc_sls}" within SLS {sls_ref} goes beyond top level package'
                )
                continue
            inc_sls = ".".join(p_comps[:-level_count] + [include])
        if inc_sls not in hub.idem.RUNS[name]["resolved"]:
            await hub.idem.sls.resolve.gather(name, inc_sls)



from typing import Any
from typing import NamedTuple


class ExecReturn(NamedTuple):
    result: bool
    ret: Any
    comment: Any = None
    ref: str = ""

    def __bool__(self):
        return self.result

    def __getitem__(self, item):
        if item in self._fields:
            return getattr(self, item)
        return tuple.__getitem__(self, item)



async def modify(hub, name, chunk):
    """
    Check the state containing the target func and call the mod_creds
    function if present. Therefore gathering the list of creds systems
    to use
    """
    run_name = chunk["ctx"]["run_name"]
    state = chunk["state"].split(".")[0]
    subs = []
    if hasattr(hub, f"states.{state}.ACCT"):
        subs = getattr(hub, f"states.{state}.ACCT")
    elif hasattr(hub, f"{state}.ACCT"):
        subs = getattr(hub, f"{state}.ACCT")
    elif hasattr(hub, f"exec.{state}.ACCT"):
        subs = getattr(hub, f"exec.{state}.ACCT")
    elif hasattr(hub, f"tool.{state}.ACCT"):
        subs = getattr(hub, f"tool.{state}.ACCT")
    hub.log.debug(f"Loaded acct from subs: {subs}")
    profile = chunk.pop("acct_profile", hub.idem.RUNS[run_name]["acct_profile"])
    hub.log.debug(f"Loaded profile: {profile}")
    chunk["ctx"]["acct"] = await hub.acct.init.gather(subs, profile)
    return chunk



def check(hub, name, ctx, condition, reqret, chunk):
    """
    Check to see if the result is True
    """
    if isinstance(condition, list):
        if reqret["ret"]["result"] in condition:
            return {}
    if reqret["ret"]["result"] is condition:
        return {}
    else:
        return {
            "errors": [
                f'Result of require {reqret["r_tag"]} is "{reqret["ret"]["result"]}", not "{condition}"'
            ]
        }



"""
This module is used to retrieve files from a json source
"""
import json
import os

__virtualname__ = "json"


async def cache(hub, cache_dir, source, loc):
    """
    Take a file from a location definition and cache it in the target location
    """
    if source.startswith("json://"):
        source = source[7:]

    data = json.loads(source)

    os.makedirs(cache_dir, exist_ok=True)

    for uuid, values in data.items():
        c_tgt = os.path.join(cache_dir, f"{uuid}.sls")
        with open(c_tgt, "w+") as fp:
            json.dump(values, fp)

        # There will only be one item in this dictionary
        return c_tgt



CLI_CONFIG = {
    # Exec Options
    "exec_func": {
        "display_priority": 0,
        "positional": True,
        "subcommands": ["exec"],
        "help": "The execution function to run by it's reference on the hub",
    },
    "exec_args": {
        "display_priority": 1,
        "positional": True,
        "nargs": "*",
        "render": "cli",
        "subcommands": ["exec"],
    },
    # State options
    "sls_sources": {"nargs": "*", "subcommands": ["state"]},
    "test": {"options": ["-t"], "action": "store_true", "subcommands": ["state"]},
    "tree": {"options": ["-T"], "subcommands": ["state"]},
    "cache_dir": {},
    "root_dir": {},
    "render": {"subcommands": ["state"]},
    "runtime": {"subcommands": ["state"]},
    "output": {"subcommands": ["exec", "state"]},
    "sls": {"positional": True, "nargs": "*", "subcommands": ["state"]},
    # ACCT options
    "acct_file": {
        "source": "acct",
        "os": "ACCT_FILE",
        "subcommands": ["state", "exec"],
    },
    "acct_key": {"source": "acct", "os": "ACCT_KEY", "subcommands": ["state", "exec"]},
    "acct_profile": {
        "os": "ACCT_PROFILE",
        "subcommands": ["state", "exec"],
    },
    "cache_plugin": {
        "os": "IDEM_CACHE_PLUGIN",
        "subcommands": ["state", "exec"],
    },
}

CONFIG = {
    "sls_sources": {
        "default": [],
        "help": "list off the sources that should be used for gathering sls files and data",
    },
    "test": {
        "default": False,
        "help": "Set the idem run to execute in test mode. No changes will be made, idem will only detect if changes will be made in a real run.",
    },
    "tree": {
        "default": "",
        "help": "The directory containing sls files",
    },
    "cache_dir": {
        "default": "/var/cache/idem",
        "help": "The location to use for the cache directory",
    },
    "root_dir": {
        "default": "/",
        "help": 'The root directory to run idem from. By default it will be "/", or in the case of running as non-root it is set to <HOMEDIR>/.idem',
    },
    "render": {
        "default": "jinja|yaml",
        "help": "The render pipe to use, this allows for the language to be specified",
    },
    "runtime": {
        "default": "parallel",
        "help": "Select which execution runtime to use",
    },
    "output": {
        "default": None,
        "help": "The outputter to use to display data",
    },
    "sls": {
        "default": [],
        "help": "A space delimited list of sls refs to execute",
    },
    "exec": {
        "default": "",
        "help": "The name of an execution function to execute",
    },
    "exec_args": {
        "default": [],
        "help": "Arguments to pass to the named execution function",
    },
    "acct_profile": {
        "os": "ACCT_PROFILE",
        "help": "The profile to use when when calling exec modules and states",
        "default": "default",
    },
    "cache_plugin": {
        "os": "IDEM_CACHE_PLUGIN",
        "subcommands": ["state", "exec"],
        "help": "The cache plugin to assist with idempotency",
        "default": "file",
    },
}
SUBCOMMANDS = {
    "state": {
        "desc": "Execute a specific state file or reference",
        "help": "Commands to execute idempotent states",
    },
    "exec": {
        "desc": "Execute a specific execution routine",
        "help": "Commands to run execution routines",
    },
}
DYNE = {
    "idem": ["idem"],
    "exec": ["exec"],
    "states": ["states"],
    "tool": ["tool"],
    "output": ["output"],
}



def check(hub, name, ctx, condition, req_ret, chunk):
    """
    If changes are made then run the configured post command
    """
    ret = {}
    if req_ret["ret"]["changes"]:
        func = getattr(hub, f'states.{chunk["state"]}.{condition}')
        if func:
            ret["post"] = func
    return ret



def post(hub, ctx):
    """
    Conform the output of every state return to this format.
    Valid state modules must return a dictionary with these keys
    """
    try:
        return {
            "changes": ctx.ret["changes"],
            "comment": ctx.ret["comment"],
            "name": ctx.ret["name"],
            "result": ctx.ret["result"],
        }
    except KeyError:
        hub.log.error(f"Improperly formatted state return: {ctx.ref}")
        raise



__func_alias__ = {"ctx_": "ctx"}


def __init__(hub):
    hub.exec.test.ACCT = ["test"]


def ping(hub):
    return {"result": True, "ret": True}


async def aping(hub):
    return {"result": True, "ret": True}


def ctx_(hub, ctx):
    return {"result": True, "comment": None, "ret": ctx}


def fail(hub):
    raise Exception("Expected failure")


async def afail(hub):
    raise Exception("Expected failure")


def echo(hub, value):
    """
    Return the parameter passed in without changing it at all
    """
    return value


async def aecho(hub, value):
    """
    Return the parameter passed in without changing it at all
    """
    return value



import asyncio


async def runtime(hub, name, ctx, seq, low, running):
    """
    Execute the runtime in parallel mode
    """
    inds = []
    for ind in seq:
        if seq[ind].get("unmet"):
            # Requisites are unmet, skip this one
            continue
        inds.append(ind)
    if not inds:
        # Nothing can be run, we have hit recursive requisite,
        # or we are done
        pass
    coros = []
    for ind in inds:
        coros.append(
            hub.idem.rules.init.run(
                name,
                ctx,
                low,
                seq[ind],
                running,
                hub.idem.RUNS[name]["run_num"],
            )
        )
        hub.idem.RUNS[name]["run_num"] += 1
    for fut in asyncio.as_completed(coros):
        await fut
    if len(low) <= len(running):
        return



"""
This module is used to retrive files from a local source
"""
import os
import shutil


__virtualname__ = "file"


async def cache(hub, cache_dir, source, loc):
    """
    Take a file from a location definition and cache it in the target location
    """
    full = os.path.join(source, loc)
    if full.startswith("file://"):
        full = full[7:]
    c_tgt = os.path.join(cache_dir, full.lstrip(os.sep))
    c_dir = os.path.dirname(c_tgt)
    os.makedirs(c_dir, exist_ok=True)
    if not os.path.isfile(full):
        return None
    shutil.copy(full, c_tgt)
    return c_tgt



import enum
import json
import tempfile
import uuid
from typing import Any
from typing import Dict

import dict_tools.update

__func_alias__ = {"compile_": "compile"}


class Status(enum.IntEnum):
    FINISHED = 0
    CREATED = 1
    GATHERING = 2
    COMPILING = 3
    RUNNING = 4
    COMPILATION_ERROR = -1
    GATHER_ERROR = -2
    RUNTIME_ERROR = -3
    UNDEFINED = -4


async def create(
    hub,
    name,
    sls_sources,
    render,
    runtime,
    subs,
    cache_dir,
    test,
    acct_file,
    acct_key,
    acct_profile,
):
    """
    Create a new instance to execute against
    """
    if acct_file and acct_key:
        await hub.acct.init.unlock(acct_file, acct_key)

    hub.idem.RUNS[name] = {
        "sls_sources": sls_sources,
        "render": render,
        "runtime": runtime,
        "subs": subs,
        "cache_dir": cache_dir,
        "states": {},
        "test": test,
        "resolved": set(),
        "files": set(),
        "high": {},
        "post_low": [],
        "errors": [],
        "iorder": 100000,
        "sls_refs": {},
        "blocks": {},
        "running": {},
        "run_num": 1,
        "add_low": [],
        "acct_profile": acct_profile,
        "status": Status.CREATED,
    }


async def apply(
    hub,
    name,
    sls_sources,
    render,
    runtime,
    subs,
    cache_dir,
    sls,
    test=False,
    acct_file=None,
    acct_key=None,
    acct_profile="default",
):
    """
    Run idem!
    """
    await hub.idem.state.create(
        name,
        sls_sources,
        render,
        runtime,
        subs,
        cache_dir,
        test,
        acct_file,
        acct_key,
        acct_profile,
    )
    # Get the sls file
    # render it
    # compile high data to "new" low data (bypass keyword issues)
    # Run the low data using act/idem
    hub.idem.RUNS[name]["status"] = Status.GATHERING
    await hub.idem.resolve.gather(name, *sls)
    if hub.idem.RUNS[name]["errors"]:
        hub.idem.RUNS[name]["status"] = Status.GATHER_ERROR
        return
    hub.idem.RUNS[name]["status"] = Status.COMPILING
    await hub.idem.state.compile(name)
    if hub.idem.RUNS[name]["errors"]:
        hub.idem.RUNS[name]["status"] = Status.COMPILATION_ERROR
        return

    hub.idem.RUNS[name]["status"] = Status.RUNNING
    try:
        await hub.idem.run.init.start(name)
    finally:
        hub.idem.RUNS[name]["status"] = Status.FINISHED


async def compile_(hub, name):
    """
    Compile the data defined in the given run name
    """
    for mod in hub.idem.compiler:
        if hasattr(mod, "stage"):
            ret = mod.stage(name)
            await hub.pop.loop.unwrap(ret)


async def single(hub, _ref_: str, _test_: bool = None, *args, **kwargs):
    """
    Run a single state and return the raw result
    :param hub:
    :param _ref_: The state's reference on the hub
    :param _test_: Run the state in a low-consequence test-mode
    :param args: Args to be passed straight through to the state
    :param kwargs: Kwargs to be passed straight through to the state
    """
    if _test_ is None:
        _test_ = hub.OPT.idem.test

    acct_file = hub.OPT.acct.acct_file
    acct_key = hub.OPT.acct.acct_key
    acct_profile = hub.OPT.acct.get("acct_profile", "default")

    args = [a for a in args]

    if not _ref_.startswith("states."):
        _ref_ = f"states.{_ref_}"

    func = getattr(hub, _ref_)
    params = func.signature.parameters

    if "ctx" in params:
        ctx = await hub.idem.ex.ctx(_ref_, acct_file, acct_key, acct_profile)
        ctx.test = _test_
        args.insert(0, ctx)

    ret = func(*args, **kwargs)
    return await hub.pop.loop.unwrap(ret)


async def batch(
    hub,
    states: Dict[str, Dict[str, Any]],
    runtime: str = None,
    test: bool = None,
    renderer: str = None,
    profiles: Dict[str, Any] = None,
    encrypted_profiles: str = None,
    acct_key: str = None,
    default_acct_profile: str = None,
    name: str = None,
):
    """
    Run multiple states defined in code
    :param hub:
    :param states: A dictionary definition of the states to run
    :param runtime: "serial" or "parallel"
    :param test: Set "test" to "True" in the implicit ctx parameter
    :param renderer: The render pipe to use
    :param profiles: An unencrypted dump of acct profiles
    :param encrypted_profiles: An encoded dump of acct profiles encrypted with a fernet key
    :param acct_key: The decryption fernet key for acct profiles, defaults to the key defined in idem's runtime config
    :param default_acct_profile: The acct profile to use for states that don't have a profile explicitly defined
    :param name: A unique identifier for this batch's run
    """
    name = name or f"name_{uuid.uuid4()}"
    runtime = runtime or hub.OPT.idem.runtime
    render = renderer or hub.OPT.idem.render
    test = hub.OPT.idem.test if test is None else test
    sls_source = f"sls_source_{uuid.uuid4()}"
    sls = [sls_source]
    data = {sls_source: states}
    sls_sources = [f"json://{json.dumps(data)}"]

    acct_key = acct_key or hub.OPT.acct.acct_key
    acct_profile = default_acct_profile or hub.OPT.acct.get("acct_profile", "default")

    if profiles:
        if not acct_key:
            # If no acct_key exists, temporarily encrypt the profiles locally with a random key
            acct_key = hub.acct.crypto.fernet.generate_key()
        elif encrypted_profiles:
            # combine the encrypted and unencrypted profiles into a single entity
            decrypted = hub.acct.crypto.fernet.decrypt(encrypted_profiles, acct_key)
            profiles = dict_tools.update.update(profiles, decrypted, merge_lists=True)

        # Encrypt the raw profiles so that idem can easily consume them
        encrypted_profiles = hub.acct.crypto.fernet.encrypt(profiles, acct_key)

    with tempfile.NamedTemporaryFile("w+") as acct_tempfile:
        if encrypted_profiles:
            acct_tempfile.write(encrypted_profiles)
            acct_file = acct_tempfile.name
        else:
            # Default to the acct_file defined in the environment at runtime
            acct_file = hub.OPT.acct.acct_file

        await hub.idem.state.apply(
            name=name,
            sls_sources=sls_sources,
            render=render,
            runtime=runtime,
            subs=["states"],
            cache_dir=hub.OPT.idem.cache_dir,
            sls=sls,
            test=test,
            acct_file=acct_file,
            acct_key=acct_key,
            acct_profile=acct_profile,
        )

    if hub.idem.RUNS[name]["errors"]:
        return hub.idem.RUNS[name]["errors"]

    return hub.idem.RUNS[name]["running"]


def status(hub, name: str) -> Dict[str, Any]:
    """
    Get the status of the named state run

        .. code-block:: json

            {
                "test": True or False,
                "errors": [],
                "running": {},
                "acct_profile": "acct profile name used for this run",
                "status": 0
                "status_name": "FINISHED"
            }
    :param hub:
    :param name: The unique identifier that was given to the state run
    :returns: A dictionary of the requested state run, An empty dictionary if the name doesn't exist
    """
    if name in hub.idem.RUNS:
        run = hub.idem.RUNS[name]
        return {
            "test": run["test"],
            "errors": run["errors"],
            "running": run["running"],
            "acct_profile": run["acct_profile"],
            "status": run["status"].value,
            "status_name": run["status"].name,
        }
    else:
        hub.log.error(f"No idem run with Job ID: {name}")
        return {
            "test": None,
            "errors": [],
            "running": {},
            "acct_profile": "",
            "status": Status.UNDEFINED.value,
            "status_name": Status.UNDEFINED.name,
        }



"""
This file contains routines to get sls files from references
"""
import os


async def ref(hub, name, sls):
    """
    Cache the given file from the named reference point
    """
    for source in hub.idem.RUNS[name]["sls_sources"]:
        proto = source[: source.index(":")]
        path = sls.replace(".", "/")
        locs = [f"{path}.sls", f"{path}/init.sls"]
        for loc in locs:
            cfn = await hub.pop.ref.last(f"idem.sls.{proto}.cache")(
                hub.idem.RUNS[name]["cache_dir"], source, loc
            )
            if cfn:
                return cfn



from typing import Any
from typing import Dict
from typing import Iterable
from typing import Tuple

import pop.hub
from dict_tools import data

__func_alias__ = {"ctx_": "ctx"}


async def run(
    hub,
    path: str,
    args: Tuple[Any],
    kwargs: Dict[str, Any],
    acct_file: str = None,
    acct_key: str = None,
    acct_profile: str = "default",
):
    args = [a for a in args]

    if not path.startswith("exec."):
        path = f"exec.{path}"

    func = getattr(hub, path)
    if isinstance(func, pop.hub.ReverseSub):
        params = func._resolve().signature.parameters
    else:
        params = func.signature.parameters

    if "ctx" in params:
        ctx = await hub.idem.ex.ctx(
            path, acct_file=acct_file, acct_key=acct_key, acct_profile=acct_profile
        )
        args.insert(0, ctx)

    ret = func(*args, **kwargs)
    return await hub.pop.loop.unwrap(ret)


async def ctx_(
    hub,
    path: str,
    acct_profile: str = "default",
    acct_file: str = None,
    acct_key: str = None,
):
    """
    :param hub:
    :param path:
    :param acct_profile:
    :param acct_file:
    :param acct_key:
    :return:
    """
    ctx = data.NamespaceDict()

    parts = path.split(".")
    try:
        parts.remove("exec")
    except ValueError:
        ...

    sname = parts[0]

    acct_paths = (f"exec.{sname}.ACCT", f"states.{sname}.ACCT")

    if acct_file and acct_key:
        await hub.acct.init.unlock(acct_file, acct_key)

    subs = set()
    for name in acct_paths:
        if hasattr(hub, name):
            sub = getattr(hub, name)
            if isinstance(sub, Iterable) and sub:
                subs.update(set(sub))

    ctx.acct = await hub.acct.init.gather(subs, acct_profile)

    return ctx


async def single(hub, path: str, *args, **kwargs):
    acct_file = hub.OPT.acct.acct_file
    acct_key = hub.OPT.acct.acct_key
    acct_profile = hub.OPT.acct.get("acct_profile", hub.acct.DEFAULT)

    ret = await hub.idem.ex.run(
        path,
        args=args,
        kwargs=kwargs,
        acct_file=acct_file,
        acct_key=acct_key,
        acct_profile=acct_profile,
    )
    return ret
